package com.example.snowm_scanner;

import android.content.Context;
import android.os.Build;
import android.util.Base64;

import androidx.annotation.RequiresApi;

import org.eclipse.paho.client.mqttv3.IMqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttCallback;
import org.eclipse.paho.client.mqttv3.MqttClient;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;

import java.io.IOException;
import java.io.InputStream;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.Properties;

import io.flutter.Log;
import io.jsonwebtoken.JwtBuilder;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureAlgorithm;

public class MqttHelper {
    private String privateKeyPath;
    private String registryId = "beacon-registry";
    private String deviceId = "BECN0IN701011111";
    private String projectId = "beaconm-c04b0";
    private String region = "us-central1";
    private String publishTopic = "/devices/" + deviceId + "/events";
    private String clientId = "projects/" + projectId + "/locations/" + region + "/registries/" + registryId + "/devices/" + deviceId;

    MqttHelper(Context context) {
        try {
            InputStream is = context.getAssets().open("rsa_private.pem");
            byte[] byteBuffer = new byte[is.available()];
            is.read(byteBuffer);
            is.close();
            privateKeyPath = new String(byteBuffer);
            privateKeyPath = privateKeyPath.replace("-----BEGIN PRIVATE KEY-----", "");
            privateKeyPath = privateKeyPath.replace("-----END PRIVATE KEY-----", "");

        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    private static String createJwtRsa(String projectId, String byteBuffer)
            throws NoSuchAlgorithmException, InvalidKeySpecException {
        final long ONE_MINUTE_IN_MILLIS = 60000;//millisecs
        Calendar d = Calendar.getInstance();
        long t = d.getTimeInMillis();
        Date now = new Date();
        Date after = new Date(t + (20 * ONE_MINUTE_IN_MILLIS));
        // Create a JWT to authenticate this device. The device will be disconnected after the token
        // expires, and will have to reconnect with a new token. The audience field should always be set
        // to the GCP project id.
        JwtBuilder jwtBuilder =
                Jwts.builder()
                        .setIssuedAt(now)
                        .setExpiration(after)
                        .setAudience(projectId);

        byte[] encoded = Base64.decode(byteBuffer, Base64.DEFAULT);
        PKCS8EncodedKeySpec spec = new PKCS8EncodedKeySpec(encoded);
        KeyFactory kf = KeyFactory.getInstance("RSA");
        return jwtBuilder.signWith(SignatureAlgorithm.RS256, kf.generatePrivate(spec)).compact();
    }

    void init(String deviceId, String message) throws MqttException, InterruptedException, NoSuchAlgorithmException, InvalidKeySpecException {
        this.deviceId = deviceId;
        MqttConnectOptions mqttConnectOptions = new MqttConnectOptions();
        mqttConnectOptions.setMqttVersion(MqttConnectOptions.MQTT_VERSION_3_1_1);

        Properties sslProps = new Properties();
        sslProps.setProperty("com.ibm.ssl.protocol", "TLSv1.2");
        mqttConnectOptions.setSSLProperties(sslProps);

        mqttConnectOptions.setUserName("unused");
        mqttConnectOptions.setPassword(createJwtRsa(projectId, privateKeyPath).toCharArray());
        mqttConnectOptions.setAutomaticReconnect(true);
        mqttConnectOptions.setCleanSession(false);

        String serverUri = "ssl://mqtt.googleapis.com:8883";
        MqttClient client = new MqttClient(serverUri, clientId, new MemoryPersistence());


        // Both connect and publish operations may fail. If they do, allow retries but with an
        // exponential backoff time period.
        long initialConnectIntervalMillis = 500L;
        long maxConnectIntervalMillis = 6000L;
        long maxConnectRetryTimeElapsedMillis = 900000L;
        float intervalMultiplier = 1.5f;

        long retryIntervalMs = initialConnectIntervalMillis;
        long totalRetryTimeMs = 0;

        while (!client.isConnected() && totalRetryTimeMs < maxConnectRetryTimeElapsedMillis) {
            try {
                client.connect(mqttConnectOptions);
                Log.d("MqttHelper", "connected to server");

            } catch (MqttException e) {
                int reason = e.getReasonCode();

                // If the connection is lost or if the server cannot be connected, allow retries, but with
                // exponential backoff.
                System.out.println("An error occurred: " + e.getMessage());
                if (reason == MqttException.REASON_CODE_CONNECTION_LOST
                        || reason == MqttException.REASON_CODE_SERVER_CONNECT_ERROR) {
                    System.out.println("Retrying in " + retryIntervalMs / 1000.0 + " seconds.");
                    Thread.sleep(retryIntervalMs);
                    totalRetryTimeMs += retryIntervalMs;
                    retryIntervalMs *= intervalMultiplier;
                    if (retryIntervalMs > maxConnectIntervalMillis) {
                        retryIntervalMs = maxConnectIntervalMillis;
                    }
                } else {
                    throw e;
                }
            }
        }
        this.client = client;
        attachCallback(client);
        publishMessage(message);

    }

    /**
     * Attaches the callback used when configuration changes occur.
     */
    private static void attachCallback(MqttClient client) {
        MqttCallback mCallback = new MqttCallback() {
            @Override
            public void connectionLost(Throwable cause) {
                Log.d("MqttHelper", "Connection was lost because of $cause");
            }

            @Override
            public void messageArrived(String topic, MqttMessage message) {
                Log.d("MqttHelper", "Incoming message :" + Arrays.toString(message.getPayload()));
            }

            @Override
            public void deliveryComplete(IMqttDeliveryToken token) {
                Log.d("MqttHelper", "Delivery completed :" + token.getResponse());
            }
        };

        client.setCallback(mCallback);
    }

    private MqttClient client;

    private void publishMessage(String _message) {

        try {
            MqttMessage message = new MqttMessage();

            message.setPayload(_message.getBytes());
            client.publish(publishTopic, message);
            Log.d("MqttHelper", "Message Published");
            if (!client.isConnected()) {
                Log.d("MqttHelper", " messages in buffer.");
            }
        } catch (MqttException e) {
            System.err.println("Error Publishing: " + e.getMessage());
            e.printStackTrace();
        }
    }


}
