package com.example.snowm_scanner

import android.content.Context
import android.os.Build
import androidx.annotation.NonNull
import androidx.annotation.RequiresApi
import io.flutter.Log
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry.Registrar
import io.flutter.embedding.engine.plugins.activity.ActivityAware;
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashMap


class SnowmScannerPlugin : FlutterPlugin, MethodCallHandler, ActivityAware {

    private lateinit var channel: MethodChannel
    private lateinit var beaconScanner: BeaconScanner
    lateinit var mqttHelper: MqttHelper
    private lateinit var utils: Utils
    val dataSentHistory: HashMap<String, Long> = HashMap()

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        channel = MethodChannel(flutterPluginBinding.getFlutterEngine().dartExecutor, "snowm_scanner")
        channel.setMethodCallHandler(this)
    }

    private fun initilizeWithContext(context: Context) {
        beaconScanner = BeaconScanner(context)
        utils = Utils(context)
        mqttHelper = MqttHelper(context)
    }

    companion object {
        @JvmStatic
        fun registerWith(registrar: Registrar) {
            val channel = MethodChannel(registrar.messenger(), "snowm_scanner")
            channel.setMethodCallHandler(SnowmScannerPlugin())
        }
    }

    @RequiresApi(Build.VERSION_CODES.O)
    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        val method: String = call.method
        when {
            method.contains("scanBeacons") -> {
                val uuids: ArrayList<String> = call.argument("uuids")!!
                val enableMqtt: Boolean = call.argument<Boolean>("enableMqtt")!!
                val timeInterval: Int = call.argument<Int>("timeInterval")!!
                if (utils.hasLocationPermission()) {
                    scan(uuids, method, enableMqtt, timeInterval)
                } else {
                    utils.requestLocationPermission()
                    scan(uuids, method, enableMqtt, timeInterval)
                }
                result.success(true)
            }
            method == "requestPermission" -> {
                result.success(utils.requestLocationPermission())
            }
            method == "permissionState" -> {
                result.success(utils.bluetoothPermissionState)
            }
            method == "bluetoothState" -> {
                result.success(utils.bluetoothState)
            }
            method.contains("bluetoothStateListener") -> {
                utils.bluetoothState(method, object : OnStateRequest {
                    override fun onState(state: String) {
                        sendBluetoothResponse(method, state)
                    }
                })
                result.success(true)
            }
            method == "cancelStream" -> {
                val streamIdentifier = call.argument<String>("methodName")!!
                beaconScanner.removeScanner(streamIdentifier)
                result.success(true)

            }
            method == "cancelBluetoothStateListener" -> {
                val streamIdentifier = call.argument<String>("methodName")!!
                utils.removeBluetoothState(streamIdentifier)
                result.success(true)

            }
            else -> {
                print("method not implemented")
                result.notImplemented()
            }
        }


    }

    private fun scan(uuids: ArrayList<String>, method: String, enableMqtt: Boolean, timeInterval: Int) {
        beaconScanner.scanBeacons(method, object : OnBeaconScanned {
            override fun onScanned(beacons: ArrayList<SnowMBeacon>) {
                val validBeacons: ArrayList<SnowMBeacon> = ArrayList()
                beacons.forEach() {
                    if (enableMqtt)
                        if (timeToSendData(dataSentHistory[it.uuid], timeInterval)) {
                            mqttHelper.init(it.uuid, it.toSendObject())
                            dataSentHistory[it.uuid] = Date().time
                        }
                    if (uuids.contains(it.uuid))
                        validBeacons.add(it)
                }

                sendBeaconResponse(method, validBeacons)
            }
        })
    }

    private fun timeToSendData(time: Long?, timeInterval: Int): Boolean {
        return if (time == null)
            true
        else {
            return Date().time.minus(time) >= timeInterval
        }
    }

    private fun sendBeaconResponse(method: String, beacons: ArrayList<SnowMBeacon>) {
        val response: HashMap<String, Any> = HashMap()
        response["beacons"] = beacons.map { b -> b.toObject() }
        channel.invokeMethod(method, response)
    }

    private fun sendBluetoothResponse(method: String, state: String) {
        val response: String = state
        channel.invokeMethod(method, response)
    }


    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        channel.setMethodCallHandler(null)
    }

    override fun onDetachedFromActivity() {
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        initilizeWithContext(binding.activity as Context)
    }

    override fun onDetachedFromActivityForConfigChanges() {
    }
}

interface OnStateRequest {
    fun onState(state: String)
}

