part of snowm_scanner;

class _SnowMScanner {
  static const MethodChannel _channel = const MethodChannel('snowm_scanner');
  int _beaconScannerCount = 0;
  int _stateScannerCount = 0;
  Map<String, StreamController<List<SnowMBeacon>>> _beaconStreamPool = {};
  Map<String, StreamController<BluetoothState>> _stateStreamPool = {};
  StreamController<BluetoothState> _stateStreamController;
  Map<String, Map<String, SnowMBeacon>> _cachedBeacons = {};

  _SnowMScanner() {
    _channel.setMethodCallHandler((call) {
      String method = call.method;
      if (method.contains("scanBeacons") &&
          _beaconStreamPool.containsKey(call.method)) {
        List<SnowMBeacon> beacons = call.arguments["beacons"]
            .map<SnowMBeacon>((b) => SnowMBeacon._fromMap(b))
            .toList();
        _beaconStreamPool[call.method].add(beacons);
      } else if (method.contains("bluetoothStateListener")) {
        _stateStreamPool[call.method]
            .add(getBluetoothStateFromKey(call.arguments));
      }
      return;
    });
  }

  Stream<BluetoothState> getBluetoothStateStream() {
    String method = "bluetoothStateListener#${++_stateScannerCount}";
    _stateStreamPool[method] =
        _getNewBluetoothStateStreamController<BluetoothState>(method);
    _channel.invokeMethod(method);
    return _stateStreamPool[method].stream;
  }

  Future<BluetoothPermissionState> getPermissionState() async {
    return getBluetoothPermissionStateFromKey(
        await _channel.invokeMethod<String>("permissionState"));
  }

  // TODO: Make it asyncronous
  void requestPermission() {
    _channel.invokeMethod<String>("requestPermission");
  }

  Future<BluetoothState> getBluetoothState() async {
    return getBluetoothStateFromKey(
        await _channel.invokeMethod<String>("bluetoothState"));
  }

  Stream<List<SnowMBeacon>> scanBeacons({
    /// List of proximity uuid of a beacon to be scanned
    List<String> uuids,

    /// Flag weather to send the data over the mqtt protocol to the cloud
    bool enableMqtt = false,

    /// Milliseconds for which the beacons should be cached. Default value is 60 seconds.
    int cacheBeaconsFor = 10000,

    /// Time interval[Milliseconds] to send data to IOT cloud.Default is 10 minutes
    /// Valid only if [enableMqtt] is true.
    int timeInterval = 600000,
  }) {
    String method = "scanBeacons#${++_beaconScannerCount}";
    _beaconStreamPool[method] =
        _getNewBeaconStreamController<List<SnowMBeacon>>(method);
    _channel.invokeMethod(method, {
      "uuids": uuids,
      "enableMqtt": enableMqtt,
      "timeInterval": timeInterval
    });
    Map<String, SnowMBeacon> cachedBeacons = _cachedBeacons[method] ?? {};
    return _beaconStreamPool[method].stream.map((beacons) {
      int currentTime = DateTime.now().millisecondsSinceEpoch;
      beacons = beacons.map((e) {
        e.._detectedTime = currentTime;
        cachedBeacons[e.uuid] = e;
        return e;
      }).toList();
      List<SnowMBeacon> freshBeacons = cachedBeacons.values
          .where((b) => (currentTime - b._detectedTime) < cacheBeaconsFor)
          .toList();
      cachedBeacons.clear();
      freshBeacons.forEach((b) {
        cachedBeacons[b.uuid] = b;
      });
      return freshBeacons;
    });
  }

  StreamController<T> _getNewBeaconStreamController<T>(String identifier) {
    StreamController<T> controller = StreamController<T>();
    controller.onCancel = () {
      _channel.invokeMethod("cancelStream", {
        "methodName": identifier,
      });
      _beaconStreamPool.remove(identifier);
    };
    return controller;
  }

  StreamController<T> _getNewBluetoothStateStreamController<T>(
      String identifier) {
    StreamController<T> controller = StreamController<T>();
    controller.onCancel = () {
      _channel.invokeMethod("cancelBluetoothStateListener", {
        "methodName": identifier,
      });
      _beaconStreamPool.remove(identifier);
    };
    return controller;
  }
}

final snowmScanner = _SnowMScanner();
